require 'helper'

describe 'store' do
  let(:store)           { Store.new({entity_class.name => entity_class}, data_mapping, entity_mapping) }
  let(:data_mapping)    {
    {
      entity_class.name => data_mapper
    }
  }
  let(:entity_mapping)  {
    {
      entity_class.name => entity_mapper
    }
  }
  let(:data_mapper)     { MiniTest::Mock.new }
  let(:entity_mapper)   { Struct.new(:store) do
    def map(entity)
      {
        'email' => entity.email,
        'gender' => entity.gender
      }
    end
    def unmap(cls, mapped)
      cls.new(mapped['email'], mapped['gender'])
    end
  end }

  let(:entity_class)    { Struct.new(:email, :gender) {def self.name; 'name'; end} }
  let(:entity)          { entity_class.new("me@foo.bar", "w") }
  let(:entity_hash)     { {'email' => entity.email, 'gender' => entity.gender} }

  let(:query)           { OpenStruct.new(:target => entity_class.name) }

  it 'delegates save to insert' do
    data_mapper.expect(:insert, nil, [entity_hash])
    store.save(entity)

    assert data_mapper.verify
  end

  it 'delegates second save to update' do
    data_mapper.expect(:insert, :ref1, [entity_hash])

    assert_equal Store::Ref.new(entity_class.name, :ref1), store.save(entity)

    data_mapper.expect(:update, :ref1, [:ref1, entity_hash])

    assert_equal  Store::Ref.new(entity_class.name, :ref1), store.save(entity)

    assert data_mapper.verify
  end

  it 'returns nil if entity of save is nil' do
    assert_nil store.save(nil)
  end

  it 'fails save if data_mapper not found' do
    entity = OpenStruct.new(:test => 42)
    error = assert_raises RuntimeError do
      store.save(entity)
    end

    assert_match(/no data_mapper .* found/i, error.message)
  end

  it 'returns references from insert/update save' do
    e1 = entity_class.new('e1', 'e1')
    e1_hash = {'email' => 'e1', 'gender' => 'e1'}
    e2 = entity_class.new('e2', 'e2')
    e2_hash = {'email' => 'e2', 'gender' => 'e2'}

    data_mapper.expect(:insert, :ref1, [e1_hash])
    store.save(e1)

    data_mapper.expect(:update, :ref1, [:ref1, e1_hash])
    data_mapper.expect(:insert, :ref2, [e2_hash])

    refs = [ 
      Store::Ref.new(entity_class.name, :ref1),
      Store::Ref.new(entity_class.name, :ref2)
    ]
    assert_equal refs, store.save([e1, e2])

    assert data_mapper.verify
  end

  it 'returns even one id as array on save with array' do
    refs = [ 
      Store::Ref.new(entity_class.name, :ref1)
    ]
    data_mapper.expect(:insert, :ref1, [entity_hash])
    assert_equal refs, store.save([entity])
  end

  it 'identifies entity as existent if changed' do
    entity.gender = 'm'

    data_mapper.expect(:insert, :ref1, [entity_hash])
    store.save(entity)

    entity.gender = 'w'

    data_mapper.expect(:update, nil, [:ref1, entity_hash.merge('gender' => 'w')])
    store.save(entity)
  end

  it 'delegates remove to delete' do
    data_mapper.expect(:insert, :ref1, [entity_hash])
    store.save(entity)

    data_mapper.expect(:delete, nil, [:ref1])
    store.remove(entity)
  end

  it 'treats entity as new after delete' do
    data_mapper.expect(:insert, :ref22, [entity_hash])
    store.save(entity)

    data_mapper.expect(:delete, nil, [:ref22])
    store.remove(entity)

    data_mapper.expect(:insert, :ref23, [entity_hash])
    store.save(entity)
  end

  it 'fails remove if data_mapper not found' do
    entity = OpenStruct.new(:test => 42)
    error = assert_raises RuntimeError do
      store.remove(entity)
    end

    assert_match(/no data_mapper .* found/i, error.message)
  end

  describe 'reload' do
    it 'delegates to load with reference' do
      data_mapper.expect(:insert, :ref22, [entity_hash])
      store.save(entity)

      data_mapper.expect(:single_find, {'email' => '1', 'gender' => '2'}, [:ref22])
      result = store.reload(entity)

      assert data_mapper.verify
      assert_equal entity_class.new('1', '2'), result
    end

    it 'raises if reference not known' do
      error = assert_raises RuntimeError do
        store.reload(entity)
      end

      assert_match(/unknown/i, error.message)
    end
  end

  describe 'load' do
    it 'accepts a ref object and delegates to single_find' do
      data_mapper.expect(:insert, :ref22, [entity_hash])
      ref = store.save(entity)

      data_mapper.expect(:single_find, entity_hash, [:ref22])
      assert_equal entity, store.load(ref)
    end

    it 'accepts an array of ref objects and delegates to bulk_find' do
      data_mapper.expect(:insert, :ref22, [entity_hash])
      ref = store.save(entity)

      data_mapper.expect(:bulk_find, [entity_hash], [[:ref22]])
      assert_equal [entity], store.load([ref])
    end

    it 'adds single_loaded entites as known entities' do
      data_mapper.expect(:insert, :ref22, [entity_hash])
      ref = store.save(entity)

      data_mapper.expect(:single_find, entity_hash, [:ref22])
      result = store.load(ref)

      data_mapper.expect(:update, :ref22, [:ref22, entity_hash])
      store.save(result)
    end

    it 'adds bulk_loaded entites as known entities' do
      data_mapper.expect(:insert, :ref22, [entity_hash])
      ref = store.save(entity)

      data_mapper.expect(:bulk_find, [entity_hash], [[:ref22]])
      result = store.load([ref])

      data_mapper.expect(:update, :ref22, [:ref22, entity_hash])
      store.save(result[0])
    end
  end

  describe 'reference' do
    it 'returns the reference for the given entity' do
      ref = Store::Ref.new('name', :ref22)
      data_mapper.expect(:insert, :ref22, [entity_hash])
      store.save(entity)

      result = store.reference(entity)
      assert_equal ref, result
    end

    it 'raises if reference not known' do
      error = assert_raises RuntimeError do
        store.reference(entity)
      end

      assert_match(/unknown/i, error.message)
    end
  end

  describe 'building' do
    Project = Struct.new(:name, :due_date)

    let(:store)               { Store.new({'Project' => Project}, {}, {}) }

    it 'should instanciate an empty object' do
      project = store.build 'Project'

      assert_nil project.name
      assert_nil project.due_date
    end

    it 'should prepopulate attributes' do
      due_date = Time.now
      project = store.build 'Project', :name => 'foo', :due_date => due_date

      assert_equal 'foo', project.name
      assert_equal due_date, project.due_date
    end
  end

  describe 'query' do
    Player = Struct.new(:email, :gender)

    let(:store)               { Store.new({'Player'=>Player}, data_mapping, entity_mapping) }
    let(:data_mapping)        do
      {
        'Player'  => player_data_mapper
      }
    end
    let(:entity_mapping)      do
      {
        'Player'  => player_entity_mapper
      }
    end
    let(:player_data_mapper)    { MiniTest::Mock.new }
    let(:player_entity_mapper)  { Struct.new(:store) do
      def map(entity)
        {
          'email' => entity.email,
          'gender' => entity.gender
        }
      end

      def unmap(cls, mapped)
        cls.new.tap do |player|
          player.email = mapped['email']
          player.gender = mapped['gender']
        end
      end
    end
    }
    let(:query) { Store::Query.new('Player', :test_query) }

    it 'delegates query to select' do
      player_data_mapper.expect(:select, nil, [query])

      store.query(query)

      assert player_data_mapper.verify
    end

    it 'returns query result' do
      e1 = Player.new('e1', 'e2')
      e1_hash = {'email' => 'e1', 'gender' => 'e2'}
      e2 = Player.new('e3', 'e4')
      e2_hash = {'email' => 'e3', 'gender' => 'e4'}

      player_data_mapper.expect(:select, {:ref1 => e1_hash, :ref2 => e2_hash}, [query])
      assert_equal [e1, e2], store.query(query)
    end

    it 'maps nil query result to empty collection' do
      player_data_mapper.expect(:select, nil, [query])
      assert_equal [], store.query(query)
    end

    it 'only calls update for queried results' do
      e1_hash = {'email' => 'e1', 'gender' => 'e1'}
      e2_hash = {'email' => 'e2', 'gender' => 'e2'}
      e3 = Player.new('e3', 'e3')
      e3_hash = {'email' => 'e3', 'gender' => 'e3'}

      player_data_mapper.expect(:select, {:ref1 => e1_hash, :ref2 => e2_hash}, [query])
      e1, e2 = store.query(query)

      player_data_mapper.expect(:update, nil, [:ref1, e1_hash])
      store.save(e1)

      player_data_mapper.expect(:update, nil, [:ref2, e2_hash])
      store.save(e2)

      player_data_mapper.expect(:insert, :ref1, [e3_hash])
      store.save(e3)

      assert player_data_mapper.verify
    end

    it 'fails if data_mapper not found' do
      query = OpenStruct.new(:entity => 'NotFound')
      error = assert_raises RuntimeError do
        store.query(query)
      end

      assert_match(/no data_mapper .* found/i, error.message)
    end
  end

  describe 'nesting' do
    User = Struct.new(:name)
    Group = Struct.new(:name, :users)

    let(:store)               { Store.new({}, data_mapping, entity_mapping) }
    let(:data_mapping)        do
      {
        'User'  => user_data_mapper,
        'Group' => group_data_mapper
      }
    end
    let(:entity_mapping)      do
      {
        'User'  => user_entity_mapper,
        'Group' => group_entity_mapper
      }
    end
    let(:user_data_mapper)    { MiniTest::Mock.new }
    let(:group_data_mapper)   { MiniTest::Mock.new }
    let(:user_entity_mapper)  { Struct.new(:store) do
      def map(entity)
        {
          'name' => entity.name
        }
      end
    end
    }
    let(:group_entity_mapper)  { Struct.new(:store) do
      def map(entity)
        {
          'name' => entity.name,
          'users' => store.save(entity.users)
        }
      end
    end
    }

    it 'should map saves for nested users' do
      users = [
        User.new('u1'),
        User.new('u2')
      ]
      group = Group.new('g1', users)

      refs = [ 
        Store::Ref.new('User', :ref1),
        Store::Ref.new('User', :ref2)
      ]
      group_data_mapper.expect :insert, nil, [{'name' => 'g1', 'users' => refs}]
      user_data_mapper.expect :insert, :ref1, [{'name' => 'u1'}]
      user_data_mapper.expect :insert, :ref2, [{'name' => 'u2'}]

      store.save(group)

      assert group_data_mapper.verify
      assert user_data_mapper.verify
    end

    it 'should map insert and update separated' do
      users = [
        User.new('u1'),
        User.new('u2')
      ]
      group = Group.new('g1', users)

      user_data_mapper.expect :insert, :uref1, [{'name' => 'u1'}]
      store.save(users[0])

      refs = [ 
        Store::Ref.new('User', :uref1),
        Store::Ref.new('User', :ref2)
      ]
      group_data_mapper.expect :insert, nil, [{'name' => 'g1', 'users' => refs}]
      user_data_mapper.expect :update, :ref1, [:uref1, {'name' => 'u1'}]
      user_data_mapper.expect :insert, :ref2, [{'name' => 'u2'}]

      store.save(group)

      assert group_data_mapper.verify
      assert user_data_mapper.verify
    end
  end
end
